class SubscribeEnvelope:
    def __init__(self, messages=None, metadata=None):
        assert isinstance(messages, (list, None))
        assert isinstance(metadata, (SubscribeMetadata, None))

        self.messages = messages
        self.metadata = metadata

    @classmethod
    def from_json(cls, json_input):
        messages = []

        for raw_message in json_input['m']:
            messages.append(SubscribeMessage.from_json(raw_message))

        metadata = SubscribeMetadata.from_json(json_input['t'])
        return SubscribeEnvelope(messages, metadata)


class SubscribeMessage:
    def __init__(self):
        self.shard = None
        self.subscription_match = None
        self.channel = None
        self.payload = None
        self.flags = None
        self.issuing_client_id = None
        self.subscribe_key = None
        self.origination_timetoken = None
        self.publish_metadata = None
        self.only_channel_subscription = False
        self.type = 0

    @classmethod
    def from_json(cls, json_input):
        message = SubscribeMessage()
        if 'a' in json_input:
            message.shard = json_input['a']
        if 'b' in json_input:
            message.subscription_match = json_input['b']
        message.channel = json_input['c']
        message.payload = json_input['d']
        message.flags = json_input['f']
        if 'i' in json_input:
            message.issuing_client_id = json_input['i']
        message.subscribe_key = json_input['k']
        if 'o' in json_input:
            message.origination_timetoken = json_input['o']
        message.publish_metadata = PublishMetadata.from_json(json_input['p'])
        if 'e' in json_input:
            message.type = json_input['e']
        return message


class SubscribeMetadata:
    def __init__(self, timetoken=None, region=None):
        self.timetoken = timetoken
        self.region = region

    @classmethod
    def from_json(cls, json_input):
        assert isinstance(json_input, dict)
        assert 'r' in json_input
        assert 't' in json_input

        return SubscribeMetadata(json_input['t'], json_input['r'])


class PresenceEnvelope:
    def __init__(self, action, uuid, occupancy, timestamp, data=None):
        assert isinstance(action, str)
        assert isinstance(occupancy, int)
        assert isinstance(timestamp, int)
        if data is not None:
            assert isinstance(data, dict)

        self.action = action
        self.uuid = uuid
        self.occupancy = occupancy
        self.timestamp = timestamp
        self.data = data

    @classmethod
    def extract_value(cls, json, key):
        if key in json:
            return json[key]
        else:
            return None

    @classmethod
    def from_json_payload(cls, json):
        return PresenceEnvelope(
            action=cls.extract_value(json, 'action'),
            uuid=cls.extract_value(json, 'uuid'),
            occupancy=cls.extract_value(json, 'occupancy'),
            timestamp=cls.extract_value(json, 'timestamp'),
            data=cls.extract_value(json, 'data')
        )


class PublishMetadata:
    def __init__(self, publish_timetoken=None, region=None):
        self.publish_timetoken = publish_timetoken
        self.region = region

    @classmethod
    def from_json(cls, json_input):
        assert 'r' in json_input
        assert 't' in json_input

        return PublishMetadata(int(json_input['t']), int(json_input['r']))



from pubnub.models.consumer.file import PNDownloadFileResult
from pubnub.endpoints.file_operations.download_file import DownloadFileNative
from pubnub.endpoints.file_operations.get_file_url import GetFileDownloadUrl


class DownloadFileAsyncio(DownloadFileNative):
    def create_response(self, envelope, data=None):
        if self._cipher_key or self._pubnub.config.cipher_key:
            data = self.decrypt_payload(data)
        return PNDownloadFileResult(data)

    async def future(self):
        self._download_data = await GetFileDownloadUrl(self._pubnub)\
            .channel(self._channel)\
            .file_name(self._file_name)\
            .file_id(self._file_id)\
            .future()

        downloaded_file = await super(DownloadFileAsyncio, self).future()
        return downloaded_file

    async def result(self):
        response_envelope = await self.future()
        return response_envelope.result



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, ListEndpoint, \
    IncludeCustomEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.channel import PNGetAllChannelMetadataResult


class GetAllChannels(ObjectsEndpoint, ListEndpoint, IncludeCustomEndpoint):
    GET_ALL_CHANNELS_PATH = "/v2/objects/%s/channels"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

    def build_path(self):
        return GetAllChannels.GET_ALL_CHANNELS_PATH % self.pubnub.config.subscribe_key

    def create_response(self, envelope):
        return PNGetAllChannelMetadataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNGetAllChannelMetadataOperation

    def name(self):
        return "Get all Channels"

    def http_method(self):
        return HttpMethod.GET



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.presence import PNHereNowResult


class HereNow(Endpoint):
    HERE_NOW_PATH = "/v2/presence/sub-key/%s/channel/%s"
    HERE_NOW_GLOBAL_PATH = "/v2/presence/sub-key/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channels = []
        self._channel_groups = []
        self._include_state = False
        self._include_uuids = True

    def channels(self, channels):
        utils.extend_list(self._channels, channels)
        return self

    def channel_groups(self, channel_groups):
        utils.extend_list(self._channel_groups, channel_groups)
        return self

    def include_state(self, should_include_state):
        self._include_state = should_include_state
        return self

    def include_uuids(self, include_uuids):
        self._include_uuids = include_uuids
        return self

    def custom_params(self):
        params = {}

        if len(self._channel_groups) > 0:
            params['channel-group'] = utils.join_items_and_encode(self._channel_groups)

        if self._include_state:
            params['state'] = "1"

        if not self._include_uuids:
            params['disable_uuids'] = "1"

        return params

    def build_path(self):
        if len(self._channels) == 0 and len(self._channel_groups) == 0:
            return HereNow.HERE_NOW_GLOBAL_PATH % self.pubnub.config.subscribe_key
        else:
            return HereNow.HERE_NOW_PATH % (self.pubnub.config.subscribe_key,
                                            utils.join_channels(self._channels))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

    def is_auth_required(self):
        return True

    def create_response(self, envelope):
        return PNHereNowResult.from_json(envelope, self._channels)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNHereNowOperation

    def name(self):
        return "HereNow"



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNGetFilesResult


class ListFiles(FileOperationEndpoint):
    LIST_FILES_URL = "/v1/files/%s/channels/%s/files"

    def __init__(self, pubnub):
        FileOperationEndpoint.__init__(self, pubnub)

    def build_path(self):
        return ListFiles.LIST_FILES_URL % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel)
        )

    def http_method(self):
        return HttpMethod.GET

    def custom_params(self):
        return {}

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()

    def create_response(self, envelope):
        return PNGetFilesResult(envelope)

    def operation_type(self):
        return PNOperationType.PNGetFilesAction

    def name(self):
        return "List files"



from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, UserEndpoint, IncludeCustomEndpoint, \
    CustomAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.entities.user import PNCreateUserResult
from pubnub.utils import write_value_as_string


class CreateUser(EntitiesEndpoint, UserEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint):
    CREATE_USER_PATH = "/v2/objects/%s/uuids/%s"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        UserEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        CustomAwareEndpoint.__init__(self)

        self._name = None
        self._email = None
        self._external_id = None
        self._profile_url = None

    def user_status(self, user_status):
        self._status = user_status
        self._include_status = True
        return self

    def user_type(self, user_type):
        self._type = user_type
        self._include_type = True
        return self

    def set_name(self, name):
        self._name = str(name)
        return self

    def email(self, email):
        self._email = str(email)
        return self

    def external_id(self, external_id):
        self._external_id = str(external_id)
        return self

    def profile_url(self, profile_url):
        self._profile_url = str(profile_url)
        return self

    def build_path(self):
        return CreateUser.CREATE_USER_PATH % (self.pubnub.config.subscribe_key, self._effective_user_id())

    def build_data(self):
        payload = {
            "name": self._name,
            "email": self._email,
            "externalId": self._external_id,
            "profileUrl": self._profile_url,
            "custom": self._custom
        }
        return write_value_as_string(payload)

    def validate_specific_params(self):
        self._validate_user_id()

    def create_response(self, envelope):
        return PNCreateUserResult(envelope)

    def operation_type(self):
        return PNOperationType.PNCreateUserOperation

    def name(self):
        return "Create User"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNFetchFileUploadS3DataResult


class FetchFileUploadS3Data(FileOperationEndpoint):
    GENERATE_FILE_UPLOAD_DATA = "/v1/files/%s/channels/%s/generate-upload-url"

    def __init__(self, pubnub):
        FileOperationEndpoint.__init__(self, pubnub)
        self._file_name = None

    def build_path(self):
        return FetchFileUploadS3Data.GENERATE_FILE_UPLOAD_DATA % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel)
        )

    def build_data(self):
        params = {
            "name": self._file_name
        }

        return utils.write_value_as_string(params)

    def http_method(self):
        return HttpMethod.POST

    def custom_params(self):
        return {}

    def is_auth_required(self):
        return True

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_name()

    def create_response(self, envelope):
        return PNFetchFileUploadS3DataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNFetchFileUploadS3DataAction

    def name(self):
        return "Fetch file upload S3 data"



from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, IncludeCustomEndpoint, ListEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.entities.space import PNFetchSpacesResult


class FetchSpaces(EntitiesEndpoint, ListEndpoint, IncludeCustomEndpoint):
    FETCH_SPACES_PATH = "/v2/objects/%s/channels"
    inclusions = ['status', 'type']

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

    def build_path(self):
        return FetchSpaces.FETCH_SPACES_PATH % self.pubnub.config.subscribe_key

    def create_response(self, envelope):
        return PNFetchSpacesResult(envelope)

    def operation_type(self):
        return PNOperationType.PNFetchSpacesOperation

    def name(self):
        return "Fetch Spaces"

    def http_method(self):
        return HttpMethod.GET



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNDeleteFileResult


class DeleteFile(FileOperationEndpoint):
    DELETE_FILE_URL = "/v1/files/%s/channels/%s/files/%s/%s"

    def __init__(self, pubnub):
        FileOperationEndpoint.__init__(self, pubnub)
        self._file_id = None
        self._file_name = None

    def build_path(self):
        return DeleteFile.DELETE_FILE_URL % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel),
            self._file_id,
            self._file_name
        )

    def file_id(self, file_id):
        self._file_id = file_id
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def http_method(self):
        return HttpMethod.DELETE

    def custom_params(self):
        return {}

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_name()
        self.validate_file_id()

    def create_response(self, envelope):
        return PNDeleteFileResult(envelope)

    def operation_type(self):
        return PNOperationType.PNDeleteFileOperation

    def name(self):
        return "Delete file"



class PNHistoryResult(object):
    def __init__(self, messages, start_timetoken, end_timetoken):
        self.messages = messages
        self.start_timetoken = start_timetoken
        self.end_timetoken = end_timetoken

    def __str__(self):
        return "History result for range %d..%d" % (self.start_timetoken, self.end_timetoken)

    @classmethod
    def from_json(cls, json_input, crypto, include_timetoken=False, include_meta=False, cipher=None):
        start_timetoken = json_input[1]
        end_timetoken = json_input[2]

        raw_items = json_input[0]
        messages = []

        for item in raw_items:
            if (include_timetoken or include_meta) and isinstance(item, dict) and 'message' in item:
                message = PNHistoryItemResult(item['message'], crypto)
                if include_timetoken and 'timetoken' in item:
                    message.timetoken = item['timetoken']
                if include_meta and 'meta' in item:
                    message.meta = item['meta']

            else:
                message = PNHistoryItemResult(item, crypto)

            if cipher is not None:
                message.decrypt(cipher)

            messages.append(message)

        return PNHistoryResult(
            messages=messages,
            start_timetoken=start_timetoken,
            end_timetoken=end_timetoken
        )


class PNHistoryItemResult(object):
    def __init__(self, entry, crypto, timetoken=None, meta=None):
        self.timetoken = timetoken
        self.meta = meta
        self.entry = entry
        self.crypto = crypto

    def __str__(self):
        return "History item with tt: %s and content: %s" % (self.timetoken, self.entry)

    def decrypt(self, cipher_key):
        self.entry = self.crypto.decrypt(cipher_key, self.entry)


class PNFetchMessagesResult(object):

    def __init__(self, channels, start_timetoken, end_timetoken):
        self.channels = channels
        self.start_timetoken = start_timetoken
        self.end_timetoken = end_timetoken

    def __str__(self):
        return "Fetch messages result for range %d..%d" % (self.start_timetoken, self.end_timetoken)

    @classmethod
    def from_json(cls, json_input, include_message_actions=False, start_timetoken=None, end_timetoken=None):
        channels = {}

        for key, entry in json_input['channels'].items():
            channels[key] = []
            for item in entry:
                message = PNFetchMessageItem(item['message'], item['timetoken'])
                if 'uuid' in item:
                    message.uuid = item['uuid']
                if 'message_type' in item:
                    message.message_type = item['message_type']

                if 'meta' in item:
                    message.meta = item['meta']

                if include_message_actions:
                    if 'actions' in item:
                        message.actions = item['actions']
                    else:
                        message.actions = {}

                channels[key].append(message)

        return PNFetchMessagesResult(
            channels=channels,
            start_timetoken=start_timetoken,
            end_timetoken=end_timetoken
        )


class PNFetchMessageItem(object):
    def __init__(self, message, timetoken, meta=None, actions=None):
        self.message = message
        self.meta = meta
        self.timetoken = timetoken
        self.actions = actions

    def __str__(self):
        return "Fetch message item with tt: %s and content: %s" % (self.timetoken, self.message)



from pubnub import utils
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.endpoints.endpoint import Endpoint


class HistoryDelete(Endpoint):  # pylint: disable=W0612
    HISTORY_DELETE_PATH = "/v3/history/sub-key/%s/channel/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._start = None
        self._end = None

    def channel(self, channel):
        self._channel = channel
        return self

    def start(self, start):
        self._start = start
        return self

    def end(self, end):
        self._end = end
        return self

    def custom_params(self):
        params = {}

        if self._start is not None:
            params['start'] = str(self._start)

        if self._end is not None:
            params['end'] = str(self._end)

        return params

    def build_path(self):
        return HistoryDelete.HISTORY_DELETE_PATH % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel)
        )

    def http_method(self):
        return HttpMethod.DELETE

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()

    def create_response(self, endpoint):
        return {}

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNHistoryDeleteOperation

    def name(self):
        return "History delete"



from pubnub.models.consumer.v3.pn_resource import PNResource


class UUID(PNResource):

    def __init__(self, resource_name=None, resource_pattern=None):
        super(UUID, self).__init__(resource_name, resource_pattern)

    @staticmethod
    def id(user_id):
        user = UUID(resource_name=user_id)
        return user

    @staticmethod
    def pattern(user_pattern):
        user = UUID(resource_pattern=user_pattern)
        return user

    def delete(self):
        self._delete = True
        return self

    def get(self):
        self._get = True
        return self

    def update(self):
        self._update = True
        return self



class PNStatus:
    def __init__(self):
        self.category = None
        self.error_data = None
        self.error = None

        self.status_code = None
        self.operation = None

        self.tls_enabled = None

        self.uuid = None
        self.auth_key = None
        self.origin = None
        self.client_request = None
        self.client_response = None
        self.original_response = None

        self.affected_channels = None
        self.affected_groups = None

    def is_error(self):
        return self.error is not None



from pubnub import utils
from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, SpaceEndpoint, SpacesEndpoint, UserEndpoint, \
    UsersEndpoint
from pubnub.enums import PNOperationType, HttpMethod
from pubnub.errors import PNERR_INVALID_SPACE, PNERR_INVALID_USER, PNERR_USER_ID_MISSING, PNERR_SPACE_MISSING
from pubnub.exceptions import PubNubException
from pubnub.models.consumer.entities.membership import PNMembershipsResult
from pubnub.models.consumer.entities.space import Space
from pubnub.models.consumer.entities.user import User


class RemoveSpaceMembers(EntitiesEndpoint, SpaceEndpoint, UsersEndpoint):
    MEMBERSHIP_PATH = "/v2/objects/%s/uuids/%s/channels"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        SpaceEndpoint.__init__(self)
        UsersEndpoint.__init__(self)

    def validate_specific_params(self):
        if self._space_id is None or len(self._space_id) == 0:
            raise PubNubException(pn_error=PNERR_SPACE_MISSING)

        self._users = list(self._users)

        if not all(isinstance(user, User) for user in self._users):
            raise PubNubException(pn_error=PNERR_INVALID_USER)

    def build_path(self):
        return RemoveSpaceMembers.MEMBERSHIP_PATH % (self.pubnub.config.subscribe_key, self.pubnub.uuid)

    def build_data(self):
        users = [user.to_payload_dict() for user in self._users]

        payload = {
            "set": [],
            "delete": users
        }
        return utils.write_value_as_string(payload)

    def create_response(self, envelope):
        return PNMembershipsResult(envelope)

    def operation_type(self):
        return PNOperationType.PNRemoveSpaceUsersOperation

    def name(self):
        return "Remove Space Users"

    def http_method(self):
        return HttpMethod.PATCH


class RemoveUserSpaces(EntitiesEndpoint, UserEndpoint, SpacesEndpoint):
    MEMBERSHIP_PATH = "/v2/objects/%s/uuids/%s/channels"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        UserEndpoint.__init__(self)
        SpacesEndpoint.__init__(self)

    def validate_specific_params(self):
        if self._user_id is None or len(self._user_id) == 0:
            raise PubNubException(pn_error=PNERR_USER_ID_MISSING)

        self._spaces = list(self._spaces)

        if not all(isinstance(space, Space) for space in self._spaces):
            raise PubNubException(pn_error=PNERR_INVALID_SPACE)

    def build_path(self):
        return RemoveUserSpaces.MEMBERSHIP_PATH % (self.pubnub.config.subscribe_key, self._user_id)

    def build_data(self):
        spaces = [space.to_payload_dict() for space in self._spaces]

        payload = {
            "set": [],
            "delete": spaces
        }
        return utils.write_value_as_string(payload)

    def create_response(self, envelope):
        return PNMembershipsResult(envelope)

    def operation_type(self):
        return PNOperationType.PNRemoveUserSpacesOperation

    def name(self):
        return "Remove User Spaces"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.errors import PNERR_UUID_MISSING
from pubnub.exceptions import PubNubException


class UUIDValidatorMixin:
    def validate_uuid(self):
        if self._uuid is None or not isinstance(self._uuid, str):
            raise PubNubException(pn_error=PNERR_UUID_MISSING)


class TimeTokenOverrideMixin:
    def replicate(self, replicate):
        self._replicate = replicate
        return self

    def ptto(self, timetoken):
        if timetoken:
            assert isinstance(timetoken, int)
            self._ptto = timetoken
        return self

    def custom_params(self):
        params = {}
        if self._replicate is not None:
            if self._replicate:
                params["norep"] = "false"
            else:
                params["norep"] = "true"

        if self._ptto:
            params["ptto"] = self._ptto

        return params



class PNErrorData():
    def __init__(self, information, exception):
        assert isinstance(information, str)
        assert isinstance(exception, Exception)

        self.information = information
        self.exception = exception



from abc import ABCMeta


class PNPage:
    __metaclass__ = ABCMeta

    def __init__(self, hash):
        self._hash = str(hash)

    @property
    def hash(self):
        return self._hash


class Next(PNPage):
    def __init__(self, hash):
        PNPage.__init__(self, hash)


class Previous(PNPage):
    def __init__(self, hash):
        PNPage.__init__(self, hash)


class PNPageable(object):
    __metaclass__ = ABCMeta

    def __init__(self, result):
        self.total_count = result.get('totalCount', None)
        if result.get("next", None):
            self.next = Next(result["next"])
        else:
            self.next = None

        if result.get("prev", None):
            self.prev = Previous(result["prev"])
        else:
            self.prev = None



from pubnub import utils
from pubnub.dtos import StateOperation
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_STATE_MISSING, PNERR_STATE_SETTER_FOR_GROUPS_NOT_SUPPORTED_YET
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.presence import PNSetStateResult


class SetState(Endpoint):
    # /v2/presence/sub-key/<subscribe_key>/channel/<channel>/uuid/<uuid>/data?state=<state>
    SET_STATE_PATH = "/v2/presence/sub-key/%s/channel/%s/uuid/%s/data"

    def __init__(self, pubnub, subscription_manager=None):
        Endpoint.__init__(self, pubnub)
        self._subscription_manager = subscription_manager
        self._channels = []
        self._groups = []
        self._state = None

    def channels(self, channels):
        utils.extend_list(self._channels, channels)
        return self

    def channel_groups(self, channel_groups):
        utils.extend_list(self._groups, channel_groups)
        return self

    def state(self, state):
        self._state = state
        return self

    def encoded_params(self):
        return {
            "state": utils.url_write(self._state)
        }

    def custom_params(self):
        if self._subscription_manager is not None:
            self._subscription_manager.adapt_state_builder(StateOperation(
                channels=self._channels,
                channel_groups=self._groups,
                state=self._state
            ))

        params = {'state': utils.write_value_as_string(self._state)}

        if len(self._groups) > 0:
            params['channel-group'] = utils.join_items_and_encode(self._groups)

        return params

    def build_path(self):
        return SetState.SET_STATE_PATH % (
            self.pubnub.config.subscribe_key,
            utils.join_channels(self._channels),
            utils.url_encode(self.pubnub.uuid)
        )

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channels_and_groups()

        if len(self._channels) == 0 and len(self._groups) > 0:
            raise PubNubException(pn_error=PNERR_STATE_SETTER_FOR_GROUPS_NOT_SUPPORTED_YET)

        if self._state is None or not isinstance(self._state, dict):
            raise PubNubException(pn_error=PNERR_STATE_MISSING)

    def create_response(self, envelope):
        if 'status' in envelope and envelope['status'] == 200:
            return PNSetStateResult(envelope['payload'])
        else:
            return envelope

    def is_auth_required(self):
        return True

    def affected_channels(self):
        return self._channels

    def affected_channels_groups(self):
        return self._groups

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNSetStateOperation

    def name(self):
        return "SetState"



import logging
import os

PUBNUB_ROOT = os.path.dirname(os.path.abspath(__file__))


def set_stream_logger(name='pubnub', level=logging.ERROR, format_string=None, stream=None):
    if format_string is None:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler(stream)
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from os import getenv
from pubnub.exceptions import PubNubException

flags = {
    'PN_ENABLE_ENTITIES': getenv('PN_ENABLE_ENTITIES', False)
}


def feature_flag(flag):
    def not_implemented(*args, **kwargs):
        raise PubNubException(errormsg='This feature is not enabled')

    def inner(method):
        if flag not in flags.keys():
            raise PubNubException(errormsg='Flag not supported')

        if not flags[flag]:
            return not_implemented
        return method
    return inner



from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, \
    ChannelEndpoint, CustomAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.channel import PNSetChannelMetadataResult


class SetChannel(ObjectsEndpoint, ChannelEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint):
    SET_CHANNEL_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self)
        CustomAwareEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

        self._name = None
        self._description = None

    def set_name(self, name):
        self._name = str(name)
        return self

    def description(self, description):
        self._description = str(description)
        return self

    def validate_specific_params(self):
        self._validate_channel()

    def build_path(self):
        return SetChannel.SET_CHANNEL_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def build_data(self):
        payload = {
            "name": self._name,
            "description": self._description,
            "custom": self._custom
        }
        return utils.write_value_as_string(payload)

    def create_response(self, envelope):
        return PNSetChannelMetadataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNSetChannelMetadataOperation

    def name(self):
        return "Set UUID"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.enums import PNOperationType, HttpMethod
from pubnub.endpoints.endpoint import Endpoint
from pubnub.models.consumer.v3.access_manager import PNRevokeTokenResult
from pubnub import utils


class RevokeToken(Endpoint):
    REVOKE_TOKEN_PATH = "/v3/pam/%s/grant/%s"

    def __init__(self, pubnub, token):
        Endpoint.__init__(self, pubnub)
        self.token = token

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_secret_key()

    def create_response(self, envelope):
        return PNRevokeTokenResult(envelope)

    def is_auth_required(self):
        return False

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def http_method(self):
        return HttpMethod.DELETE

    def custom_params(self):
        return {}

    def build_path(self):
        return RevokeToken.REVOKE_TOKEN_PATH % (self.pubnub.config.subscribe_key, utils.url_encode(self.token))

    def operation_type(self):
        return PNOperationType.PNAccessManagerRevokeToken

    def name(self):
        return "RevokeToken"



from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_CHANNEL_MISSING, PNERR_PUSH_DEVICE_MISSING, PNERROR_PUSH_TYPE_MISSING, \
    PNERR_PUSH_TOPIC_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType, PNPushType, PNPushEnvironment
from pubnub.models.consumer.push import PNPushRemoveChannelResult
from pubnub import utils


class RemoveChannelsFromPush(Endpoint):
    # v1/push/sub-key/{subKey}/devices/{pushToken}
    REMOVE_PATH = "/v1/push/sub-key/%s/devices/%s"
    # v2/push/sub-key/{subKey}/devices-apns2/{deviceApns2}
    REMOVE_PATH_APNS2 = "/v2/push/sub-key/%s/devices-apns2/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channels = None
        self._device_id = None
        self._push_type = None
        self._topic = None
        self._environment = None

    def channels(self, channels):
        self._channels = channels
        return self

    def device_id(self, device_id):
        self._device_id = device_id
        return self

    def push_type(self, push_type):
        self._push_type = push_type
        return self

    def topic(self, topic):
        self._topic = topic
        return self

    def environment(self, environment):
        self._environment = environment
        return self

    def custom_params(self):
        params = {'remove': utils.join_items(self._channels)}

        if self._push_type != PNPushType.APNS2:
            params['type'] = utils.push_type_to_string(self._push_type)
        else:
            if self._environment is None:
                self._environment = PNPushEnvironment.DEVELOPMENT

            params['environment'] = self._environment
            params['topic'] = self._topic

        return params

    def build_path(self):
        if self._push_type != PNPushType.APNS2:
            return RemoveChannelsFromPush.REMOVE_PATH % (
                self.pubnub.config.subscribe_key, self._device_id)
        else:
            return RemoveChannelsFromPush.REMOVE_PATH_APNS2 % (
                self.pubnub.config.subscribe_key, self._device_id)

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()

        if not isinstance(self._channels, list) or len(self._channels) == 0:
            raise PubNubException(pn_error=PNERR_CHANNEL_MISSING)

        if not isinstance(self._device_id, str) or len(self._device_id) == 0:
            raise PubNubException(pn_error=PNERR_PUSH_DEVICE_MISSING)

        if self._push_type is None:
            raise PubNubException(pn_error=PNERROR_PUSH_TYPE_MISSING)

        if self._push_type == PNPushType.APNS2:
            if not isinstance(self._topic, str) or len(self._topic) == 0:
                raise PubNubException(pn_error=PNERR_PUSH_TOPIC_MISSING)

    def create_response(self, envelope):
        return PNPushRemoveChannelResult()

    def is_auth_required(self):
        return True

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNRemovePushNotificationsFromChannelsOperation

    def name(self):
        return "RemoveChannelsFromPush"



from pubnub.models.consumer.objects_v2.page import PNPageable


class PNSetUUIDMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Set UUID metatdata: %s" % self.data


class PNGetUUIDMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get UUID metatdata: %s" % self.data


class PNRemoveUUIDMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get UUID metatdata: %s" % self.data


class PNGetAllUUIDMetadataResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get all UUID metatdata: %s" % self.data


class PNUUIDMetadataResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "UUID %s event with data: %s" % (self.event, self.data)



from typing import Optional
from pubnub.models.consumer.entities.result import PNEntityPageableResult, PNEntityResult


class PNCreateSpaceResult(PNEntityResult):
    _description = "Create Space: %s"


class PNUpdateSpaceResult(PNEntityResult):
    _description = "Update Space: %s"


class PNFetchSpaceResult(PNEntityResult):
    _description = "Fetch Space: %s"


class PNRemoveSpaceResult(PNEntityResult):
    _description = "Remove Space: %s"


class PNFetchSpacesResult(PNEntityPageableResult):
    _description = "Fetch Spaces: %s"


class PNSpaceResult(PNEntityResult):
    def __str__(self):
        return "Space %s event with data: %s" % (self.event, self.data)


class Space:
    space_id: str
    custom: Optional[dict]

    def __init__(self, space_id=None, **kwargs):
        self.space_id = space_id
        if 'custom' in kwargs.keys():
            self.custom = kwargs['custom']

    def to_payload_dict(self):
        result = {
            "channel": {
                "id": str(self.space_id)
            }
        }
        if 'custom' in self.__dict__.keys():
            result['custom'] = self.custom
        return result



from abc import abstractmethod, ABCMeta

from pubnub.models.consumer.objects_v2.page import PNPageable


class PNUUID:
    __metaclass__ = ABCMeta

    def __init__(self, uuid):
        self._uuid = uuid

    @staticmethod
    def uuid(uuid):
        return JustUUID(uuid)

    @staticmethod
    def uuid_with_custom(uuid, custom):
        return UUIDWithCustom(uuid, custom)

    @abstractmethod
    def to_payload_dict(self):
        return None


class JustUUID(PNUUID):
    def to_payload_dict(self):
        return {
            "uuid": {
                "id": str(self._uuid)
            }
        }


class UUIDWithCustom(PNUUID):
    def __init__(self, uuid, custom):
        PNUUID.__init__(self, uuid)
        self._custom = custom

    def to_payload_dict(self):
        return {
            "uuid": {
                "id": str(self._uuid)
            },
            "custom": dict(self._custom)
        }


class PNSetChannelMembersResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Set Channel Members metatdata: %s" % self.data


class PNGetChannelMembersResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get Channel Members metatdata: %s" % self.data


class PNRemoveChannelMembersResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Remove Channel Members metatdata: %s" % self.data


class PNManageChannelMembersResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Manage Channel Members metatdata: %s" % self.data



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.errors import PNERR_TTL_MISSING, PNERR_INVALID_META, PNERR_RESOURCES_MISSING
from pubnub.exceptions import PubNubException
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.v3.access_manager import PNGrantTokenResult


class GrantToken(Endpoint):
    GRANT_TOKEN_PATH = "/v3/pam/%s/grant"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._ttl = None
        self._meta = None
        self._authorized_uuid = None
        self._channels = []
        self._groups = []
        self._uuids = []

        self._sort_params = True

    def ttl(self, ttl):
        self._ttl = ttl
        return self

    def meta(self, meta):
        self._meta = meta
        return self

    def authorized_uuid(self, uuid):
        self._authorized_uuid = uuid
        return self

    def authorized_user(self, user):
        self._authorized_uuid = user
        return self

    def spaces(self, spaces):
        self._channels = spaces
        return self

    def users(self, users):
        self._uuids = users
        return self

    def channels(self, channels):
        self._channels = channels
        return self

    def groups(self, groups):
        self._groups = groups
        return self

    def uuids(self, uuids):
        self._uuids = uuids
        return self

    def custom_params(self):
        return {}

    def build_data(self):
        params = {'ttl': int(self._ttl)}

        permissions = {}
        resources = {}
        patterns = {}

        utils.parse_resources(self._channels, "channels", resources, patterns)
        utils.parse_resources(self._groups, "groups", resources, patterns)
        utils.parse_resources(self._uuids, "uuids", resources, patterns)
        utils.parse_resources(self._uuids, "users", resources, patterns)
        utils.parse_resources(self._channels, "spaces", resources, patterns)

        permissions['resources'] = resources
        permissions['patterns'] = patterns

        if self._meta:
            if isinstance(self._meta, dict):
                permissions['meta'] = self._meta
            else:
                raise PubNubException(pn_error=PNERR_INVALID_META)
        else:
            permissions['meta'] = {}

        if self._authorized_uuid:
            permissions["uuid"] = self._authorized_uuid

        params['permissions'] = permissions

        return utils.write_value_as_string(params)

    def build_path(self):
        return GrantToken.GRANT_TOKEN_PATH % self.pubnub.config.subscribe_key

    def http_method(self):
        return HttpMethod.POST

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_secret_key()
        self.validate_ttl()
        self.validate_resources()

    def create_response(self, envelope):
        return PNGrantTokenResult.from_json(envelope['data'])

    def is_auth_required(self):
        return False

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNAccessManagerGrantToken

    def name(self):
        return "Grant Token"

    def validate_resources(self):
        if not any((self._channels, self._groups, self._uuids)):
            raise PubNubException(pn_error=PNERR_RESOURCES_MISSING)

    def validate_ttl(self):
        if not self._ttl:
            raise PubNubException(pn_error=PNERR_TTL_MISSING)



from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, IncludeCustomEndpoint, \
    ChannelEndpoint, ListEndpoint, UUIDIncludeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.channel_members import PNGetChannelMembersResult


class GetChannelMembers(ObjectsEndpoint, ChannelEndpoint, ListEndpoint, IncludeCustomEndpoint,
                        UUIDIncludeEndpoint):
    GET_CHANNEL_MEMBERS_PATH = "/v2/objects/%s/channels/%s/uuids"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        ChannelEndpoint.__init__(self)
        ListEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        UUIDIncludeEndpoint.__init__(self)

    def build_path(self):
        return GetChannelMembers.GET_CHANNEL_MEMBERS_PATH % (self.pubnub.config.subscribe_key, self._channel)

    def validate_specific_params(self):
        self._validate_channel()

    def create_response(self, envelope):
        return PNGetChannelMembersResult(envelope)

    def operation_type(self):
        return PNOperationType.PNGetChannelMembersOperation

    def name(self):
        return "Get Channel Members"

    def http_method(self):
        return HttpMethod.GET



class PNSignalResult(object):
    def __init__(self, result):
        """
        Representation of signal server response

        :param result: result of signal operation
        """
        self.timetoken = result[2]
        self._result = result

    def __str__(self):
        return "Signal success with timetoken %s" % self.timetoken
